import pyfunk.combinators as _


def fid(a):
    '''fid :: a -> a'''
    return a


def fconstant(a):
    '''fconstant :: a -> (* -> a)'''
    return lambda *args, **kwargs: a


@_.curry
def trace(tag, val):
    '''trace :: Show a => String -> a -> a'''
    print(tag, val)
    return val


def T(*args):
    '''
    Always returns true
    @sig T :: _ -> Bool
    '''
    return True


def F(*args):
    '''
    Always returns false
    @sig F :: _ -> Bool
    '''
    return False



from pyfunk.combinators import curry


class Monad(object):

    def __init__(self, val):
        self._value = val

    @classmethod
    def of(cls, val):
        '''
        Creates a new applicative
        @sig of :: Applicative a => x -> a x
        '''
        return cls(val)

    def fmap(self, fn):
        '''
        Transforms the value of the functor using the given function
        @sig fmap :: Functor f => (a -> b) -> f b
        '''
        return self.of(fn(self._value))

    def join(self):
        '''
        Lifts a functor out of another
        @sig join :: Functor f => f (f a) -> f a
        '''
        return self._value

    def chain(self, fn):
        '''
        Applies a function that produces a chain to the value in this
        chain
        @sig chain :: Chain c => (a -> c b) -> c b
        '''
        return self.fmap(fn).join()

    def ap(self, f):
        '''
        Applies the function in the Apply on the given functor's
        value. This function must be curried
        @sig ap :: (Apply a,Functor f) => f x -> a y
        '''
        return f.fmap(self._value)

    @curry
    def liftA2(self, f1, f2):
        '''
        Applies the function in the apply to 2 functors
        @sig ap :: (Apply a,Functor f) => f x -> f y -> a z
        '''
        return f1.fmap(self._value).ap(f2)

    @curry
    def liftA3(self, f1, f2, f3):
        '''
        Applies the function in the apply to 2 functors
        @sig ap :: (Apply a,Functor f) => f x -> f x -> f y -> a z
        '''
        return f1.fmap(self._value).ap(f2).ap(f3)



from pyfunk.monads import Monad
from pyfunk.combinators import curry


def ftry(fn):
    '''
    Equivalent to try-except using Either monad
    @sig ftry :: Either e => (* -> a) -> (* -> e b a)
    '''
    def trier(*args, **kwargs):
        try:
            Right.of(fn(*args, **kwargs))
        except Exception as e:
            Left.of(e)
    return trier


@curry
def cata(f, g, e):
    '''
    Extracts the values from either for transformation there f is for
    Left and g for Right
    @sig cata :: Either a b => (a -> c) -> (b -> c) -> Either a b -> c
    '''
    if isinstance(e, Left):
        return f(e._value)
    elif isinstance(e, Right):
        return g(e._value)


class Left(Monad):

    def fmap(self, fn):
        '''
        Does nothing
        @sig fmap :: Left l => (a -> b) -> l a
        '''
        return self

    def join(self):
        '''
        Returns the same left monad
        @sig join :: Left l => _ -> l a
        '''
        return self


class Right(Monad):
    pass



from pyfunk.monads import Monad


class Writer(Monad):

    def __init__(self, val, *info):
        super().__init__(val)
        self._info = info

    @classmethod
    def of(cls, val):
        '''
        Unit function for creating writer monad
        @sig of :: Writer w => a -> w a
        '''
        return Writer(val)

    def fmap(self, fn):
        '''
        Transforms the value in the Writer monad producing
        a new value and extra info
        @sig fmap :: Writer w => (a -> (b c)) -> w b
        '''
        val, info = fn(self._value)
        return Writer(val, *(self.info + (info,)))



from functools import reduce, wraps


def curry(f):
    '''
    Returns the curried equivalent of the given function.
    @sig curry :: * -> b -> * -> b
    '''
    @wraps(f)
    def curried(*args, **kwargs):
        if len(args) + len(kwargs) == f.__code__.co_argcount:
            return f(*args, **kwargs)
        return lambda *args2, **kwargs2: curried(*(args + args2), **dict(kwargs, **kwargs2))
    return curried


def compose(*fns):
    '''
    Performs right-to-left function composition. The rightmost function may have
    any arity, the remaining functions must be unary.
    @sig compose :: (b -> c)...(* -> b) -> (* -> c)
    '''
    return reduce(lambda f, g:
                  lambda *args: f(g(*args)), fns)


def fnot(f):
    '''
    Creates a function that negates the result of the predicate.
    @sig fnot :: (* -> Bool) -> * -> Bool
    '''
    return lambda *args: not f(*args)


@curry
def chain(fn, f):
    '''
    Chains a function to a functor using its bound api
    @sig chain :: Functor f => (a -> b) -> f a -> f b
    '''
    return f.chain(fn)



from pyfunk.combinators import compose
from pyfunk.monads import Monad


class Task(Monad):

    def __init__(self, fn):
        '''
        Create new Task
        @sig a -> Task a b
        '''
        self.fork = fn

    @classmethod
    def of(cls, x):
        '''
        Factory for creating new resolved task
        @sig of :: b -> Task _ b
        '''
        return Task(lambda _, resolve: resolve(x))

    @classmethod
    def rejected(cls, x):
        '''
        Factory for creating a rejected task
        @sig rejected :: a -> Task a _
        '''
        return Task(lambda reject, _: reject(x))

    def fmap(self, fn):
        '''
        Transforms the resolved value of the task using the given function
        @sig fmap :: Task a b => (b -> c) -> Task a c
        '''
        return Task(lambda rej, res: self.fork(rej, compose(res, fn)))

    def rejected_fmap(self, fn):
        '''
        Transforms the rejected value of the task using the given function
        @sig rejected_fmap :: Task a b => (a -> c) -> Task c b
        '''
        return Task(lambda rej, res: self.fork(compose(rej, fn), res))

    def join(self):
        '''
        Lifts a Task out of another
        @sig join :: Task a b => Task a Task a b -> Task a b
        '''
        return Task(lambda rej, res: self.fork(rej,
                    lambda x: x.fork(rej, res)))

    def or_else(self, fn):
        '''
        Transforms a failure value into a new `Task[α, β]`. Does nothing if the
        structure already contains a successful value.
        @sig chain :: Task a b => (a -> Task c) -> Task c b
        '''
        return Task(lambda rej, res: self.fork(lambda x: fn(x).fork(rej, res), res))

    def freduce(self, fn, x):
        '''
        [WIP]
        Equivalent of reduce for arrays
        @sig freduce :: Task t => _ -> _ -> a
        '''
        pass



from pyfunk import combinators as _
from pyfunk.misc import T
from functools import reduce


def __map_object(obj, fn):
    def reducer(newObj, key):
        newObj[key] = fn(obj[key])
        return newObj
    return reduce(reducer, obj.keys(), {})


def __filter_object(obj, fn):
    def reducer(newObj, key):
        if fn(obj[key]):
            newObj[key] = obj[key]
        return newObj
    return reduce(reducer, obj.keys(), {})


@_.curry
def fmap(fn, arr):
    '''
    Generic function for dealing with objects and arrays.
    @sig fmap :: (a -> b) -> [a] -> [b]
    '''
    if hasattr(arr, 'keys'):
        return __map_object(arr, fn)
    else:
        return list(map(fn, arr))


@_.curry
def fslice(x, y, arr):
    '''
    Composable equivalent of [:]. Use None for the second argument
    for the equivalent [x:]
    @sig fslice :: Number -> Number -> [a] -> [a]
    '''
    return arr[x:y]


@_.curry
def freduce(fn, init, arr):
    '''
    Composable equivalent of reduce in functional tools
    @sig freduce :: (a -> b) -> a -> [a] -> b
    '''
    return reduce(fn, arr, init)


@_.curry
def ffilter(fn, arr):
    '''
    Composable equivalent of iterable filter function.
    @sig ffilter :: (a -> Bool) -> [a] -> [a]
    '''
    if hasattr(arr, 'keys'):
        return __filter_object(arr, fn)
    else:
        return list(filter(fn, arr))


@_.curry
def prop(string, obj):
    '''
    Property access as a function
    @sig prop :: String -> Dict -> a
    '''
    return obj.get(string)


@_.curry
def concat(c1, c2):
    '''
    Equivalent of + for concatables(string, array)
    @sig concat :: Concatable c => c a -> c a -> c a
    '''
    return c1 + c2


@_.curry
def index_of(fn, arr):
    '''
    Find the index of first element that passes given.
    @sig index_of :: (a -> Bool) -> [a] -> Number
    '''
    for i in range(len(arr)):
        if fn(arr[i]):
            return i
    return -1


@_.curry
def array_get(i, arr):
    '''
    Array access as a function.
    @sig array_get :: Number -> [a] ->  a
    '''
    try:
        return arr[i]
    except IndexError:
        return None


@_.curry
def first(fn, arr):
    '''
    Get the first element that passes the test.
    @sig first :: (a -> Bool) -> [a] -> a
    '''
    return array_get(index_of(fn, arr), arr)

# Get the top of array
# head :: [a] -> a
head = first(T)


@_.curry
def take_while(fn, arr):
    '''
    Retrieve the first set of elements that pass the given test.
    @sig take_while :: (a -> Bool) -> [a] -> [a]
    '''
    i = index_of(_.fnot(fn), arr)
    return fslice(0, 0 if i == -1 else i, arr)


@_.curry
def drop_while(fn, arr):
    '''
    Slice array starting from the element that fails the given test.
    @sig drop_while :: (a -> Bool) -> [a] -> [a]
    '''
    i = index_of(_.fnot(fn), arr)
    return fslice(len(arr) if i == -1 else i, None, arr)



from pyfunk.combinators import compose, curry
from pyfunk.collections import fmap as cfmap


@curry
def fmap(fn, f):
    '''
    Generic version of fmap.
    @sig fmap :: Functor f => (a -> b) -> f a -> f b
    '''
    return f.fmap(fn) if hasattr(f, 'fmap') else cfmap(fn, f)


@curry
def chain(fn, c):
    '''
    Generic version of chain
    @sig chain :: Chain c => (a -> c b) -> c a -> c b
    '''
    return c.chain(fn)


@curry
def ap(fof, fn, f):
    '''
    Generic ap
    @sig ap :: Applicative a, Functor f => (t -> a t) -> (x -> y) -> f x -> a y
    '''
    return fof(fn).ap(f)


def ccompose(*fns):
    '''
    Composes functions that produce Chains
    @sig mcompose :: Chain c => (y -> c z)...(x -> c y) -> (x -> c z)
    '''
    last = fns[-1:]
    rest = tuple(list(map(chain, fns[:-1])))
    chained = rest + last
    return compose(*chained)


@curry
def liftA2(fn, f1, f2):
    '''
    Generic version of liftA2
    @sig ap :: Functor f => (x -> y -> z) -> f x -> f y -> a z
    '''
    return f1.fmap(fn).ap(f2)


@curry
def liftA3(fn, f1, f2, f3):
    '''
    Generic version of liftA3
    @sig ap :: Functor f => (w -> x -> y -> z) f v -> f x -> f y -> a z
    '''
    return f1.fmap(fn).ap(f2).ap(f3)


from pyfunk.monads.base import Monad  # noqa
from pyfunk.monads.identity import Identity  # noqa
from pyfunk.monads.const import Const  # noqa
from pyfunk.monads.either import Left, Right, ftry, cata  # noqa
from pyfunk.monads.io import IO  # noqa
from pyfunk.monads.maybe import Maybe, maybe, maybify  # noqa
from pyfunk.monads.task import Task  # noqa



from pyfunk.monads import Monad
from pyfunk.combinators import compose


class IO(Monad):

    def __init__(self, fn):
        '''
        Create new IO Monad
        @sig a -> IO a
        '''
        self.io = fn

    @classmethod
    def of(cls, x):
        '''
        Factory for creating new IO monad
        @sig of :: a -> IO a
        '''
        return cls(lambda: x)

    def fmap(self, fn):
        '''
        Transforms the value of the IO monad using the given function
        @sig fmap :: IO a => (a -> b) -> IO b
        '''
        return IO(compose(fn, self.io))

    def join(self):
        '''
        Lifts an IO monad out of another
        @sig join :: IO i => i (i a) -> c a
        '''
        return IO(lambda: self.io().io())



from pyfunk.monads import Monad


class Const(Monad):

    def fmap(self, fn):
        '''
        Ignores the given function
        @sig fmap :: Functor f => _ -> f a
        '''
        return self

    def join(self):
        '''
        Returns itself
        @sig join :: Functor f => f (f a) -> f a
        '''
        return self



"""Set functional dev tools"""
from pip.req import parse_requirements
from setuptools import setup, find_packages


dependencies = parse_requirements('requirements.txt', session=False)
reqs = [str(ir.req) for ir in dependencies]
setup(
    name='gakp-pyfunk',
    version='0.1.2',
    url='https://gitlab.com/danceasarxx/gakp-pyfunc.git',
    author='Arewa Olakunle',
    author_email='arewa.olakunle@gmail.com',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=reqs,
    long_description=open("README.md").read()
)



from pyfunk.monads import Monad
from pyfunk.combinators import curry, compose


@curry
def maybe(x, fn, mb):
    '''
    Extract a Maybe's value providing a default if Nothing
    @sig maybe :: b -> (a -> b) -> Maybe a -> b
    '''
    return x if mb.nothing() else fn(mb._value)


@curry
def maybify(nullify, f):
    '''
    Creates a function that returns Maybe. Also accepts a Nonifier
    that checks that transforms falsy value to None. One can use
    misc.F if there is no need to nonify.
    @sig maybify :: (a -> Bool) -> (a -> b) -> (a -> Maybe b)
    '''
    return compose(lambda a: Maybe.of(None if nullify(a) else a), f)


class Maybe(Monad):

    def nothing(self):
        '''
        Checks if the value of this Maybe is None.
        @sig nothing :: Maybe m => _ -> Bool
        '''
        return self._value is None

    def fmap(self, fn):
        '''
        Transforms the value of the Maybe monad using the given function
        @sig fmap :: Maybe m => (a -> b) -> Maybe b
        '''
        return (Maybe.of(None) if self.nothing()
                else Maybe.of(fn(self._value)))

    def join(self):
        '''
        Lifts an Maybe monad out of another
        @sig join :: Maybe i => i (i a) -> c a
        '''
        return (Maybe.of(None) if self.nothing()
                else self._value)

    def or_else(self, x):
        '''
        Returns the default if the value does not
        exist
        @sig or_else :: Maybe m => x -> a
        '''
        return x if self.nothing() else self._value
